// disable zooming
window.addEventListener('keydown', (event) => {
  if ((event.ctrlKey || event.metaKey) && ['+', '-', '0'].includes(event.key)) {
    event.preventDefault();
  }
});


//
// async init()
// Async iife to check if the code trainer at the specified location is available.
(async function init() {
  // get the main content area
  const $content = document.querySelector('footer');

  // signal the user that we are starting checking the availability
  $content.innerHTML = `
    <p>Prüfe Code Trainer Verfügbarkeit...</p>
  `;

  // determine the stored or default array of code trainer urls
  const servers = JSON.parse(localStorage.getItem('servers') ?? null) ?? [{
    url: 'https://codetrainer.app/',
    default: true,
  }];

  // determine if the app is running in an development environment
  const isDev = await (async () => {
    try {
      await require('fs/promises').access(`${process.resourcesPath}/.git/`);
      return true;
    } catch {
      return false;
    }
  })();

  // check if we shall launch the app in server selection mode
  const manual = await new Promise((resolve) => {
    window.addEventListener('keydown', ({ key }) => {
      if (key === 'b') {
        if (localStorage.getItem('always_boot_menu')) {
          localStorage.removeItem('always_boot_menu');
          return true;
        } else {
          localStorage.setItem('always_boot_menu', 'yes');
          return true;
        }
      } else {
        resolve(key === 'Control');
      }
    });
    setTimeout(() => resolve(localStorage.getItem('always_boot_menu')), 1000);
  });

  // use the default server or show the server selection list, if the selection mode has been
  // started and wait for the user to select a server manually
  const url = !manual ? servers.find((server) => server.default).url : (await new Promise((resolve) => {
    // empty the content and add a new empty list
    $content.innerHTML = '<p>Server auswählen:</p>';
    const $list = $content.appendChild(document.createElement('ul'));

    //
    // interaction()
    // Helper method to handle interaction with the servers list item
    // this = Object<url: String, default: Boolean, ref: Element>
    // <- Object<target: Element>
    function interaction({ target }) {
      // provide a better name for the this bound server object
      const server = this;

      // handle clicks on the images
      if (target.tagName === 'IMG') {
        // star icon to mark the server as default
        if (target.src.includes('star')) {
          // mark all servers as non default
          servers.forEach((server) => {
            server.default = false;
            server.ref.querySelector('img').src = 'star.svg';
          });

          // mark the clicked server as default
          server.default = true;
          server.ref.querySelector('img').src = 'star.filled.svg';
        }

        // delete button to delete a non default server
        else if (target.src.includes('delete')) {
          // skip the default server, as this one is not allowed to be deleted
          if (server.default) return;

          // delete the server and remove it from the list
          servers.splice(servers.indexOf(server), 1);
          server.ref.remove();
        }

        // store the current list of servers to the local storage
        localStorage.setItem('servers', JSON.stringify(servers));

        // done
        return;
      }

      // ending up here means the user clicked somewhere else on the entry thus selecting it to be
      // used as the code trainer server of choice
      resolve(server.url);
    };

    // iterate over the list of servers and add the localhost server as well, if we are in a dev
    // environment
    for (const server of [isDev ? [{ url: 'http://localhost:8080/' }] : [], ...servers].flat()) {
      // build and add the wrapper for the server in the ui list
      const $server = $list.appendChild(document.createElement('li'));

      // add a reference to the server wrapper to the server storage object
      server.ref = $server;

      // insert the contents into the wrapper, but spare the editing options from it, if it is not
      // included in the list of servers, which is the case for the local test server in a develop
      // environment
      $server.innerHTML = `
        <span>${server.url}</span>
        ${servers.includes(server) === false ? '' : `
          <hr>
          <img src="${server.default ? 'star.filled.svg' : 'star.svg'}">
          <img src="delete.svg">
        `}
      `;

      // add interactivity to the wrapper and bind the server storage object to it
      $server.addEventListener('click', interaction.bind(server));
    }

    // handle the creation and features of the new server button at the bottom of the list
    {
      // build and add the wrapper of the create button as well as insert its contents
      const $new = $content.appendChild(document.createElement('div'));
      $new.innerHTML = `
        <img src="add.svg">
      `;

      // add interactivity to the creation button, so when it gets clicked it shall add a new, yet
      // empty, server entry to the ui list
      $new.addEventListener('click', () => {
        // build and add the wrapper for the new server in the ui list
        const $server = $list.appendChild(document.createElement('li'));

        // add an editable span as well as the other default contents to the wrapper
        $server.innerHTML = `
          <span contenteditable>https://</span>
          <hr>
          <img src="star.svg">
          <img src="delete.svg">
        `;

        // handle url editing in the new server entry, which gets focussed right away and will store
        // or delete the entry based on the edited content on blur
        {
          // access the span right away
          const $url = $server.querySelector('span');

          // set the focus on the span and put the cursor at the end
          $url.focus();
          window.getSelection().getRangeAt(0).setStartAfter($url.firstChild);

          // blurring the span means checking if the url is basically set and storing it or throwing
          // it away
          $url.addEventListener('blur', () => {
            // access the current text of the url and remove any whitespace at the ends as well as
            // newlines, in case the user pasted something weird into the field
            let url = $url.textContent.trim().replaceAll('\n', '');

            // check if we can create an url object, which means the user entered something valid we
            // can use, else we show an error and drop it
            try {
              url = new URL(url);
              url = url.href;
              $url.textContent = url;
            } catch {
              alert(`Die Adresse '${url}' ist ungültig und kann nicht verwendet werden.`);
              $server.remove();
              return;
            }

            // after the url gets focussed no further editing is possible. this also shows the star
            // and delete buttons for the first time on the servers ui entry
            $url.toggleAttribute('contenteditable', false);

            // create the completed new server entry
            const server = {
              url,
              default: false,
              ref: $server,
            };

            // store the current list of servers and within the local storage
            servers.push(server);
            localStorage.setItem('servers', JSON.stringify(servers));

            // add the interactivity, making the new server entry in the ui complete and fully
            // functional
            $server.addEventListener('click', interaction.bind(server));
          });

          // using the enter key shall trigger a blur of the span, so no newline will get inserted
          $url.addEventListener('keydown', (event) => {
            if (event.key !== 'Enter') return;
            event.preventDefault();
            $url.blur();
          });
        }
      });
    }
  }));

  // ensure the checking server label is shown, which is not the case, when we have been in the
  // manual server selection mode
  $content.innerHTML = `
    <p>Prüfe Code Trainer Verfügbarkeit...</p>
  `;

  // safely check if the code trainer is available
  try {
    // build the proper path to the destination, as we require the server to serve a specific
    // endpoint to prove it is a code trainer server
    const destination = new URL(url);
    destination.pathname = `/api/training/available`;

    // use a head request to check if the server of the code trainer answers
    const test = await fetch(destination, { method: 'HEAD' });

    // now check if the result is positive with the specific status code 202 and without a text,
    // which means the requested location does host the code trainer, in which case we forward to it
    if (test.status === 202 && await test.text() === '') {
      location.href = url;
      return;
    }
  }

  // catch and log any errors during the availability check
  catch(error) {
    console.error(error);
  }

  // if we reach this point, then the availability check was not successful and we need to inform
  // the user about that issue and provide a reload button
  $content.innerHTML = `
    <p>Code Trainer ist nicht verfügbar.</p>
    <p>Stelle sicher, dass eine Verbindung zum Server <a href="${url}" target="_blank">${url}</a> besteht.</p>
    <div>
      <img src="reload.svg">
    </div>
  `;

  // soft reload the page when the reload button gets clicked
  $content.querySelector('img').parentNode.addEventListener('click', init);
})();
